#include "fft.h"
#include "./kiss_fft/tools/kiss_fftr.h"
#include <math.h>


// {2C943F6E-ADAB-4e0d-BE4F-72F1078E8BDD}
const GUID spectrum::guid = 
{ 0x2c943f6e, 0xadab, 0x4e0d, { 0xbe, 0x4f, 0x72, 0xf1, 0x7, 0x8e, 0x8b, 0xdd } };

spectrum * spectrum::create(const char * name) {
	service_enum_t< spectrum > e;
	spectrum * ptr;
	for (ptr = e.first(); ptr; ptr = e.next()) {
		if (!stricmp_utf8(name, ptr->get_name()))
			return ptr;
		ptr->service_release();
	}
	return 0;
}

static unsigned next_power_of_two(unsigned n) {
	unsigned rv = 1;
	while (rv && rv < n)
		rv = rv << 1;
	return rv;
}

#ifndef PI
#define PI 3.141592653589793238462643383279502884197169399375
#endif

static vis_sample blackman(unsigned idx, unsigned len) {
	double p = (double)idx / double(len-1);
	return (vis_sample) (0.42 - 0.5 * cos(2*PI*p) + 0.8 * cos(4*PI*p));
}

static vis_sample hanning(unsigned idx, unsigned len) {
	double inv = 1 / (double)len;
	return (vis_sample) sqrt(cos((PI*inv)*(idx-(double)(len-1)/2)));
}

class spectrum_plain : public spectrum {
private:
	unsigned samples;
	unsigned padded_size;
	mem_block_t< kiss_fft_scalar > padded_in;
	mem_block_t< kiss_fft_cpx > padded_out;
	void * kiss_cfg;
public:
	spectrum_plain() {
		samples = 0;
		padded_size = 0;
		kiss_cfg = 0;
	}

	~spectrum_plain() {
		if (kiss_cfg) kiss_fft_free(kiss_cfg);
	}

	virtual const char * get_name() {
		return "Plain";
	}

	virtual bool run(const vis_chunk * in, vis_chunk * out) {
		if (in->samples != samples) {
			samples = in->samples;
			if (samples > 256) samples = 256;
			int zpf = 1;
			padded_size = next_power_of_two(in->samples) * zpf;
			padded_in.check_size(padded_size);
			padded_out.check_size(padded_size);
			if (kiss_cfg) kiss_fft_free(kiss_cfg);
			kiss_cfg = kiss_fftr_alloc(padded_size, 0, 0, 0);
		}
		unsigned n;
		vis_sample * ptr = in->data;
		for (n = 0; n < samples; n++) {
			padded_in[n] = (kiss_fft_scalar) (*ptr * blackman(n, samples));
			ptr += in->nch;
		}
		for (n = samples; n < padded_size; n++) {
			padded_in[n] = 0;
		}
		kiss_fftr(kiss_cfg, padded_in, padded_out);
		unsigned out_size = 256;
		out->samples = in->samples;
		out->flags = in->flags;
		out->srate = in->srate;
		out->data = 0;
		out->nch = 1;
		out->spectrum_size = out_size;
		out->spectrum = mem_ops< vis_sample >::alloc(out_size);
		if (padded_size/2 <= out_size) {
			unsigned f = out_size * 2 / padded_size;
			for (n = 0; n < out_size; n++) {
				unsigned idx = n / f;
				out->spectrum[n] = (vis_sample) sqrt( padded_out[idx].r * padded_out[idx].r + padded_out[idx].i * padded_out[idx].i);
			}
		} else {
			unsigned f = padded_size / 2 / out_size;
			unsigned m;
			for (n = 0; n < out_size; n++) {
				vis_sample v = 0;
				for (m = 0; m < f; m++) {
					int idx = n * f + m;
					v = v + (vis_sample) sqrt( padded_out[idx].r * padded_out[idx].r + padded_out[idx].i * padded_out[idx].i);
				}
				out->spectrum[n] = v / (vis_sample)f;
			}
		}
		return true;
	}
};

static spectrum_factory< spectrum_plain > foo;
